@@ -31,8 +31,8 @@ explicit method call.
 
 Example:
     # Wrap two large data sources (e.g. 52000 x 800 x 600).
-    measured = ArrayAdapter(netcdf_var_a)
-    predicted = ArrayAdapter(netcdf_var_b)
+    measured = OrthoArrayAdapter(netcdf_var_a)
+    predicted = OrthoArrayAdapter(netcdf_var_b)
 
     # No actual calculations are performed here.
     error = predicted - measured
@@ -59,7 +59,7 @@ import numpy as np
 import numpy.ma as ma
 
 
-__version__ = '0.2'
+__version__ = '0.3'
 
 
 class Array(object):
@@ -70,10 +70,29 @@ class Array(object):
     """
     __metaclass__ = ABCMeta
 
+    @staticmethod
+    def ndarrays(arrays):
+        """
+        Return a list of NumPy ndarray objects corresponding to the given
+        biggus Array objects.
+
+        Subclasses may override this method to provide more efficient
+        implementations for their instances.
+
+        """
+        return [array.ndarray() for array in arrays]
+
+    __hash__ = None
+
     def __repr__(self):
         return '<{} shape={} dtype={!r}>'.format(type(self).__name__,
                                                  self.shape, self.dtype)
 
+    @property
+    def fill_value(self):
+        """The value used to fill in masked values where necessary."""
+        return np.ma.empty(0, dtype=self.dtype).fill_value
+
     @property
     def ndim(self):
         """The number of dimensions in this virtual array."""
@@ -108,17 +127,17 @@ class Array(object):
         """
 
 
-class ArrayAdapter(Array):
+class _ArrayAdapter(Array):
     """
-    Exposes a "concrete" array (e.g. numpy.ndarray, netCDF4.Variable)
-    as an Array.
+    Abstract base class for exposing a "concrete" data source as a
+    :class:`biggus.Array`.
 
     """
     def __init__(self, concrete, keys=()):
         # concrete must have:
         #   dtype
         #   shape
-        self._concrete = concrete
+        self.concrete = concrete
         if not isinstance(keys, tuple):
             keys = (keys,)
         assert len(keys) <= len(concrete.shape)
@@ -130,11 +149,18 @@ class ArrayAdapter(Array):
 
     @property
     def dtype(self):
-        return self._concrete.dtype
+        return self.concrete.dtype
+
+    @property
+    def fill_value(self):
+        fill_value = getattr(self.concrete, 'fill_value', None)
+        if fill_value is None:
+            fill_value = Array.fill_value.fget(self)
+        return fill_value
 
     @property
     def shape(self):
-        shape = _sliced_shape(self._concrete.shape, self._keys)
+        shape = _sliced_shape(self.concrete.shape, self._keys)
         return shape
 
     def _cleanup_new_key(self, key, size, axis):
@@ -205,7 +231,7 @@ class ArrayAdapter(Array):
             raise IndexError('too many keys')
 
         result_keys = []
-        shape = list(self._concrete.shape)
+        shape = list(self.concrete.shape)
         src_keys = list(self._keys or [])
         new_keys = list(keys)
 
@@ -247,8 +273,44 @@ class ArrayAdapter(Array):
                 result_keys.append(result_key)
                 axis += 1
 
-        return ArrayAdapter(self._concrete, tuple(result_keys))
+        return type(self)(self.concrete, tuple(result_keys))
 
+    @abstractmethod
+    def _apply_keys(self):
+        pass
+
+    def ndarray(self):
+        array = self._apply_keys()
+        # We want the shape of the result to match the shape of the
+        # Array, so where we've ended up with an array-scalar,
+        # "inflate" it back to a 0-dimensional array.
+        if array.ndim == 0:
+            array = np.array(array)
+        if ma.isMaskedArray(array):
+            array = array.filled()
+        return array
+
+    def masked_array(self):
+        array = self._apply_keys()
+        # We want the shape of the result to match the shape of the
+        # Array, so where we've ended up with an array-scalar,
+        # "inflate" it back to a 0-dimensional array.
+        if array.ndim == 0 or not ma.isMaskedArray(array):
+            array = ma.MaskedArray(array, fill_value=self.fill_value)
+        return array
+
+
+class NumpyArrayAdapter(_ArrayAdapter):
+    """
+    Exposes a "concrete" data source which supports NumPy "fancy
+    indexing" as a :class:`biggus.Array`.
+
+    A NumPy ndarray instance is an example suitable data source.
+
+    NB. NumPy "fancy indexing" contrasts with orthogonal indexing which
+    treats multiple iterable index keys as independent.
+
+    """
     def _apply_keys(self):
         # If we have more than one tuple as a key, then NumPy does
         # "fancy" indexing, instead of "column-based" indexing, so we
@@ -275,38 +337,46 @@ class ArrayAdapter(Array):
                 cut_keys = list(keys)
                 for i, key in tuple_keys:
                     cut_keys[i] = slice(None)
-                array = self._concrete[tuple(cut_keys)]
+                array = self.concrete[tuple(cut_keys)]
                 is_scalar = [isinstance(key, int) for key in cut_keys]
                 dimensions -= np.cumsum(is_scalar)
             else:
-                array = self._concrete
+                # Use ellipsis indexing to ensure we have a real ndarray
+                # instance to work with. (Otherwise self.concrete would
+                # need to implement `take` or `__array__`.)
+                array = self.concrete[...]
             # ... and then do each tuple in turn.
             for i, key in tuple_keys:
-                cut_keys = [slice(None)] * dimensions[i]
-                cut_keys.append(key)
-                array = array[tuple(cut_keys)]
+                array = np.take(array, key, axis=dimensions[i])
         else:
-            array = self._concrete.__getitem__(keys)
+            array = self.concrete.__getitem__(keys)
         return array
 
-    def ndarray(self):
-        array = self._apply_keys()
-        # We want the shape of the result to match the shape of the
-        # Array, so where we've ended up with an array-scalar,
-        # "inflate" it back to a 0-dimensional array.
-        if array.ndim == 0:
-            array = np.array(array)
-        if ma.isMaskedArray(array):
-            array = array.filled()
-        return array
 
-    def masked_array(self):
-        array = self._apply_keys()
-        # We want the shape of the result to match the shape of the
-        # Array, so where we've ended up with an array-scalar,
-        # "inflate" it back to a 0-dimensional array.
-        if array.ndim == 0 or not ma.isMaskedArray(array):
-            array = ma.MaskedArray(array)
+class OrthoArrayAdapter(_ArrayAdapter):
+    """
+    Exposes a "concrete" data source which supports orthogonal indexing
+    as a :class:`biggus.Array`.
+
+    Orthogonal indexing treats multiple iterable index keys as
+    independent (which is also the behaviour of a :class:`biggus.Array`).
+
+    For example::
+
+        >>> ortho_concrete.shape
+        (100, 200, 300)
+        >> ortho_concrete[(0, 3, 4), :, (1, 9)].shape
+        (3, 200, 2)
+
+    A netCDF4.Variable instance is an example orthogonal concrete array.
+
+    NB. Orthogonal indexing contrasts with NumPy "fancy indexing" where
+    multiple iterable index keys are zipped together to allow the
+    selection of sparse locations.
+
+    """
+    def _apply_keys(self):
+        array = self.concrete.__getitem__(self._keys)
         return array
 
 
@@ -319,17 +389,24 @@ class ArrayStack(Array):
         first_array = stack.flat[0]
         item_shape = first_array.shape
         dtype = first_array.dtype
+        fill_value = first_array.fill_value
         for array in stack.flat:
-            if array.shape != item_shape or array.dtype != dtype:
+            if (array.shape != item_shape or array.dtype != dtype or
+                    array.fill_value != fill_value):
                 raise ValueError('invalid sub-array')
         self._stack = stack
         self._item_shape = item_shape
         self._dtype = dtype
+        self._fill_value = fill_value
 
     @property
     def dtype(self):
         return self._dtype
 
+    @property
+    def fill_value(self):
+        return self._fill_value
+
     @property
     def shape(self):
         return self._stack.shape + self._item_shape
@@ -377,11 +454,11 @@ class ArrayStack(Array):
         return data
 
     def masked_array(self):
-        data = ma.empty(self.shape, dtype=self.dtype)
+        data = ma.empty(self.shape, dtype=self.dtype,
+                        fill_value=self.fill_value)
         for index in np.ndindex(self._stack.shape):
             masked_array = self._stack[index].masked_array()
             data[index] = masked_array
-            data.fill_value = masked_array.fill_value
         return data
 
 
@@ -397,6 +474,7 @@ class LinearMosaic(Array):
         # Make sure all the tiles are compatible
         common_shape = list(first.shape)
         common_dtype = first.dtype
+        common_fill_value = first.fill_value
         del common_shape[axis]
         for tile in tiles[1:]:
             shape = list(tile.shape)
@@ -405,19 +483,28 @@ class LinearMosaic(Array):
                 raise ValueError('inconsistent tile shapes')
             if tile.dtype != common_dtype:
                 raise ValueError('inconsistent tile dtypes')
+            if tile.fill_value != common_fill_value:
+                raise ValueError('inconsistent tile fill_values')
         self._tiles = tiles
         self._axis = axis
+        self._cached_shape = None
 
     @property
     def dtype(self):
         return self._tiles[0].dtype
 
+    @property
+    def fill_value(self):
+        return self._tiles[0].fill_value
+
     @property
     def shape(self):
-        shape = list(self._tiles[0].shape)
-        for tile in self._tiles[1:]:
-            shape[self._axis] += tile.shape[self._axis]
-        return tuple(shape)
+        if self._cached_shape is None:
+            shape = list(self._tiles[0].shape)
+            for tile in self._tiles[1:]:
+                shape[self._axis] += tile.shape[self._axis]
+            self._cached_shape = tuple(shape)
+        return self._cached_shape
 
     def __getitem__(self, keys):
         if not isinstance(keys, tuple):
@@ -500,7 +587,8 @@ class LinearMosaic(Array):
         return data
 
     def masked_array(self):
-        data = ma.empty(self.shape, dtype=self.dtype)
+        data = ma.empty(self.shape, dtype=self.dtype,
+                        fill_value=self.fill_value)
         offset = 0
         indices = [slice(None)] * self.ndim
         axis = self._axis
@@ -521,40 +609,71 @@ def ndarrays(arrays):
     individual arrays one by one.
 
     """
-    # Pick out any _Aggregation nodes which have a simple ArrayAdapter
-    # source.
-    ndarrays = [None] * len(arrays)
-    aggregations_by_src = {}
+    # Group the given Arrays by their static ndarrays() methods.
+    index_array_pairs_by_func = {}
     for i, array in enumerate(arrays):
-        if isinstance(array, _Aggregation) and \
-                isinstance(array._array, ArrayAdapter):
-            aggregations = aggregations_by_src.setdefault(array._array, [])
-            aggregations.append((i, array))
+        index_array_pairs = index_array_pairs_by_func.setdefault(
+            array.ndarrays, [])
+        index_array_pairs.append((i, array))
+    # Call each static ndarrays() method and compile the results.
+    all_results = [None] * len(arrays)
+    for func, index_array_pairs in index_array_pairs_by_func.iteritems():
+        indices = [index for index, array in index_array_pairs]
+        results = func([array for index, array in index_array_pairs])
+        for i, ndarray in zip(indices, results):
+            all_results[i] = ndarray
+    return all_results
+
+
+MAX_CHUNK_SIZE = 1024 * 1024
+
+
+def _all_slices(array):
+    # Return the slices for each dimension which ensure complete
+    # coverage by chunks no larger than MAX_CHUNK_SIZE.
+    # e.g. For a float32 array of shape (100, 768, 1024) the slices are:
+    #   (0, 1, 2, ..., 99),
+    #   (slice(0, 256), slice(256, 512), slice(512, 768)),
+    #   (slice(None)
+    nbytes = array.dtype.itemsize
+    all_slices = []
+    for i, size in reversed(list(enumerate(array.shape))):
+        if size * nbytes <= MAX_CHUNK_SIZE:
+            slices = (slice(None),)
+        elif nbytes > MAX_CHUNK_SIZE:
+            slices = range(size)
         else:
-            ndarrays[i] = array.ndarray()
-    for aggregations in aggregations_by_src.itervalues():
-        indexes = [agg[0] for agg in aggregations]
-        agg_arrays = [agg[1] for agg in aggregations]
-        agg_ndarrays = _aggregation_ndarrays(agg_arrays)
-        for i, ndarray in zip(indexes, agg_ndarrays):
-            ndarrays[i] = ndarray
-    return ndarrays
+            step = MAX_CHUNK_SIZE / nbytes
+            slices = []
+            for start in range(0, size, step):
+                slices.append(slice(start, start + step))
+        nbytes *= size
+        all_slices.insert(0, slices)
+    return all_slices
 
 
-def _aggregation_ndarrays(arrays):
-    chunk_handlers = [array.chunk_handler() for array in arrays]
-    for chunk_handler in chunk_handlers:
-        chunk_handler.bootstrap()
+def save(sources, targets):
+    """
+    Save the numeric results of each source into its corresponding target.
 
-    def meta_chunk_handler(chunk):
-        for chunk_handler in chunk_handlers:
-            chunk_handler.add_chunk(chunk)
+    """
+    # TODO: Remove restriction
+    assert len(sources) == 1 and len(targets) == 1
+    array = sources[0]
+    target = targets[0]
 
-    src_array = arrays[0]._array
-    _process_chunks(src_array, meta_chunk_handler)
-
-    results = [chunk_handler.result() for chunk_handler in chunk_handlers]
-    return results
+    # Request bitesize pieces of the source and assign them to the
+    # target.
+    # NB. This algorithm does not use the minimal number of chunks.
+    #   e.g. If the second dimension could be sliced as 0:99, 99:100
+    #       then clearly the first dimension would have to be single
+    #       slices for the 0:99 case, but could be bigger slices for the
+    #       99:100 case.
+    # It's not yet clear if this really matters.
+    all_slices = _all_slices(array)
+    for index in np.ndindex(*[len(slices) for slices in all_slices]):
+        keys = tuple(slices[i] for slices, i in zip(all_slices, index))
+        target[keys] = array[keys].ndarray()
 
 
 class _ChunkHandler(object):
@@ -580,7 +699,7 @@ class _ChunkHandler(object):
 
 class _Mean(_ChunkHandler):
     def bootstrap(self):
-        first_slice = self.array[self.axis].ndarray()
+        first_slice = self.array[0].ndarray()
         self.running_total = np.array(first_slice)
         self.t = np.empty_like(first_slice)
 
@@ -602,7 +721,7 @@ class _Std(_ChunkHandler):
     # http://zach.in.tu-clausthal.de/teaching/info_literatur/Welford.pdf
 
     def bootstrap(self):
-        first_slice = self.array[self.axis].ndarray()
+        first_slice = self.array[0].ndarray()
         self.a = np.array(first_slice)
         self.q = np.zeros_like(first_slice)
         self.t = np.empty_like(first_slice)
@@ -631,6 +750,46 @@ class _Std(_ChunkHandler):
 
 
 class _Aggregation(Array):
+    @staticmethod
+    def ndarrays(arrays):
+        """
+        Return a list of NumPy ndarray objects corresponding to the given
+        biggus _Aggregation objects.
+
+        """
+        assert all(isinstance(array, _Aggregation) for array in arrays)
+
+        # Group the given Arrays by their sources.
+        index_array_pairs_by_source = {}
+        for i, array in enumerate(arrays):
+            index_array_pairs = index_array_pairs_by_source.setdefault(
+                id(array._array), [])
+            index_array_pairs.append((i, array))
+        all_results = [None] * len(arrays)
+        for index_array_pairs in index_array_pairs_by_source.itervalues():
+            indices = [index for index, array in index_array_pairs]
+            arrays = [array for index, array in index_array_pairs]
+            results = _Aggregation._ndarrays_common_source(arrays)
+            for i, ndarray in zip(indices, results):
+                all_results[i] = ndarray
+        return all_results
+
+    @staticmethod
+    def _ndarrays_common_source(arrays):
+        chunk_handlers = [array.chunk_handler() for array in arrays]
+        for chunk_handler in chunk_handlers:
+            chunk_handler.bootstrap()
+
+        def meta_chunk_handler(chunk):
+            for chunk_handler in chunk_handlers:
+                chunk_handler.add_chunk(chunk)
+
+        src_array = arrays[0]._array
+        _process_chunks(src_array, meta_chunk_handler)
+
+        results = [chunk_handler.result() for chunk_handler in chunk_handlers]
+        return results
+
     def __init__(self, array, axis, chunk_handler_class, kwargs):
         self._array = array
         self._axis = axis
@@ -656,7 +815,10 @@ class _Aggregation(Array):
                             self._chunk_handler_class, self._kwargs)
 
     def ndarray(self):
-        return ndarrays([self])[0]
+        chunk_handler = self.chunk_handler()
+        chunk_handler.bootstrap()
+        _process_chunks(self._array, chunk_handler.add_chunk)
+        return chunk_handler.result()
 
     def masked_array(self):
         raise RuntimeError()
@@ -672,6 +834,7 @@ def mean(a, axis=None):
     NB. Currently limited to axis=0.
 
     """
+    assert axis == 0
     return _Aggregation(a, axis, _Mean, {})
 
 
@@ -682,9 +845,66 @@ def std(a, axis=None, ddof=0):
     NB. Currently limited to axis=0.
 
     """
+    assert axis == 0
     return _Aggregation(a, axis, _Std, {'ddof': ddof})
 
 
+class _Elementwise(Array):
+    def __init__(self, array1, array2, numpy_op, ma_op):
+        # TODO: Broadcasting
+        assert array1.shape == array2.shape
+        # TODO: Type-promotion
+        assert array1.dtype == array2.dtype
+        self._array1 = array1
+        self._array2 = array2
+        self._numpy_op = numpy_op
+        self._ma_op = ma_op
+
+    @property
+    def dtype(self):
+        return self._array1.dtype
+
+    @property
+    def shape(self):
+        return self._array1.shape
+
+    def __getitem__(self, keys):
+        if not isinstance(keys, tuple):
+            keys = (keys,)
+        return _Elementwise(self._array1[keys], self._array2[keys],
+                            self._numpy_op, self._ma_op)
+
+    def _calc(self, op):
+        operands = (self._array1, self._array2)
+        np_operands = ndarrays(operands)
+        result = op(*np_operands)
+        return result
+
+    def ndarray(self):
+        result = self._calc(self._numpy_op)
+        return result
+
+    def masked_array(self):
+        result = self._calc(self._ma_op)
+        return result
+
+
+def add(a, b):
+    """
+    Return the elementwise evaluation of `a + b` as another Array.
+
+    """
+    return _Elementwise(a, b, np.add, np.ma.add)
+
+
+def sub(a, b):
+    """
+    Return the elementwise evaluation of `a - b` as another Array.
+
+    """
+    return _Elementwise(a, b, np.subtract, np.ma.subtract)
+
+
 def _process_chunks(array, chunk_handler):
     #   chunk_size = 2      => 54s ~ 115% CPU
     #   chunk_size = 10     => 42s ~ 105% CPU (quicker than CDO!)
@@ -711,6 +931,16 @@ def _process_chunks(array, chunk_handler):
     chunks.join()
 
 
+def _process_chunks_simple(array, chunk_handler):
+    # Simple, single-threaded version for debugging.
+    size = array.shape[0]
+    chunk_size = 10
+
+    for i in range(1, size, chunk_size):
+        chunk = array[i:i + chunk_size].ndarray()
+        chunk_handler(chunk)
+
+
 # TODO: Test
 def _sliced_shape(shape, keys):
     """
